# -*- coding: utf-8 -*-
"""
This module contains the JSON schema used to validate various inputs and outputs.
"""


## This could crank down harder, but it should be enough to be confident we are looking at geojson.
geojson_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title":"In-built ArcGIS",
    
    "type":"object",
    "properties":{
        "type":{"type":"string", "enum":["Feature", "FeatureCollection"]},
        "features":{"type":"array", 
                    "items":{"type":"object",
                             "properties":{
                                 "type":{"type":"string", "enum":["Feature"]},
                                 "properties":{"type":["object", "null"]},
                                 "geometry":{"type":["null", "object"]}},
                             "required":["type", "properties", "geometry"]}}},
    "required":["type"],
    "if":{"properties":{"type":{"const":"Feature"}}},
    "then":{
        "additionalProperties":{
            "type":"object",
            "properties":{
                "type":{"type":"string", "enum":["Feature"]},
                "properties":{"type":["object", "null"]},
                "geometry":{"type":["null", "object"]}},
            "required":["type", "properties", "geometry"]}}}

geojson_feature_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title":"In-built GEOJSON Single Feature",
    
    "type":"object",
    "properties":{
        "type":{"type":"string", "enum":["Feature"]},
        "properties":{"type":["object", "null"]},
        "geometry":{"type":["null", "object"]}},
    "required":["type", "properties", "geometry"]
    }

geojson_collection_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title":"In-built GEOJSON Collection",
    
    "type":"object",
    "properties":{
        "type":{"type":"string", "enum":["FeatureCollection"]},
        "features":{"type":"array", 
                    "items":{"type":"object",
                             "properties":{
                                 "type":{"type":"string", "enum":["Feature"]},
                                 "properties":{"type":["object", "null"]},
                                 "geometry":{"type":["null", "object"]}},
                             "required":["type", "properties", "geometry"]}}},
    "required":["type", "features"],
    }

## This is an abbreviated schema that should be enough to check if the JSON is the format we are looking for.
arcgis_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title":"In-built ESRI",
    
    "type":"object",
    "properties":{
        "layers":{"type":"array", 
                  "items":{"type":"object",
                           "properties":{
                               "layerDefinition":{"type":"object",
                                                  "properties":{
                                                      "fields":{"type":"array",
                                                                "items":{"type":"object",
                                                                         "properties":{
                                                                             "name":{"type":"string"},
                                                                             "type":{"type":"string"}},
                                                                         "required":["name","type"]}}},
                                                  "required":["fields"]}},
                           "required":["layerDefinition"]}},},
    "required":["layers"]}




metadata_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    
    "type":"object",
    "properties":{
        "format_version":{"type":"string", "minLength":1},
        "entry_version":{"type":"integer", "minimum":1},
        "entry_id":{"type":"string", "minLength":1},
        "date":{"type":"string", "minLength":1},
        "description":{"type":"string", "minLength":1},
        "products":{"type":"array", "items":{"type":"string", "minLength":1}},
        "resources":{"type":"object",
                     "minProperties":1,
                     "additionalProperties":{"type":"object",
                                             "properties":{
                                                 "location":{"type":"string", "minLength":1},
                                                 "alternate_locations":{"type":"array", "items":{"type":"string", "minLength":1}, "minItems":1},
                                                 "type":{"type":"string", "minLength":1},
                                                 "fairness":{"type":"string", "pattern":"(?i)^f?a?i?r?$"},
                                                 "format":{"type":"string", "minLength":1},
                                                 "schema":{"type":["string", "object"], "minLength":1},
                                                 "sources":{"type":"array", "items":{"type":"string", "minLength":1}},
                                                 "description":{"type":"string", "minLength":1},
                                                 "creator":{"type":"array", "minItems":1, "items":{"type":"object",
                                                                                                   "properties":{
                                                                                                       "name":{"type":"string", "minLength":1},
                                                                                                       "type":{"type":"string", "minLength":1, "enum":["URL", "organization", "author", "DOI"]}},
                                                                                                   "required":["name", "type"]}},
                                                 "fields":{"type":"object",
                                                           "minProperties":1,
                                                           "additionalProperties":{"type":"object",
                                                                                   "properties":{
                                                                                       "name":{"type":"string", "minLength":1},
                                                                                       "type":{"type":"string", "enum":["ontology_term", "int", "float", "str"]},
                                                                                       "identifier":{"type":["string", "integer"], "minLength":1},
                                                                                       "identifier%type":{"type":"string", "minLength":1}},
                                                                                   "if":{"anyOf":[
                                                                                       {"properties":{"identifier":{"type":["string", "integer"], "minLength":1}},
                                                                                        "required":["identifier"]},
                                                                                       {"properties":{"identifier%type":{"type":"string", "minLength":1}},
                                                                                        "required":["identifier%type"]}]},
                                                                                   "then":{"required":["name", "type", "identifier", "identifier%type"]},
                                                                                   "else":{"required":["name", "type"]}}}},
                                             "if":{"anyOf":[
                                                     {"properties":{"type":{"anyOf":[{"const":"program"},
                                                                                     {"const":"other"}]}}},
                                                     {"properties":{"format":{"const":"web"}}}]},
                                             "then":{"required":["location", "type", "fairness", "format", "description"]},
                                             "else":{"required":["location", "type", "fairness", "format", "description", "fields"]}}},
        },
    "required":["format_version", "entry_version", "entry_id", "date", "description", "products", "resources"]}



args_schema = {
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "CLI inputs",
 
 "type":"object",
 "properties":{
     "--resource_properties":{"type":["string", "null"], "minLength":1},
     "--json_schemas":{"type":["string", "null"], "minLength":1},
     "--entry_id":{"type":["string", "null"], "minLength":1},
     "--description":{"type":["string", "null"], "minLength":1},
     "--base_metadata":{"type":["string", "null"], "minLength":1}},
 }




schema_list_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "JSON Schemas List",
    
    "type":"array",
    "items":{"type":"object",
             "properties":{
                 "name":{"type":"string", "minLength":1},
                 "style":{"type":"string", "enum":["testing", "mapping"]},
                 "schema":{"type":"object"},
                 "schema_URL":{"type":"string", "minLength":1},
                 
                 "field_path":{"type":"string", "minLength":1},
                 "name_key":{"type":"string", "minLength":1},
                 "type_key":{"type":"string", "minLength":1},
                 "type_map":{"type":"object", "minProperties":1},
                 
                 "features_path":{"type":"string"},
                 "properties_key":{"type":"string", "minLength":1}},
             "required":["name", "schema", "style"],
             "allOF":[
                 {"if":{"properties":{"style":{"const":"mapping"}}},
                 "then":{"required":["field_path", "name_key", "type_key", "type_map"]},},
                 {"if":{"properties":{"style":{"const":"testing"}}},
                 "then":{"required":["features_path", "properties_key"]}}
                 ]
             }
    
    }












# -*- coding: utf-8 -*-
"""
This module contains the code to validate the metadata file.
"""

import re

import jsonschema

from . import miagis_schema

def validate(metadata: dict):
    """Validate input metadata against the MIAGIS schema.
    
    Use jsonschema to validate input metadata, and then do some additional checking 
    that jsonschema alone cannot, such as checking that all layers for each map 
    exist in the layers section.
    
    The specific JSON schema used is the metadata_schema in the miagis_schema module.
    
    Args:
        metadata: input dictionary of metadata.
        
    Raises:
        jsonschema.ValidationError: any validation errors that aren't handled reraise the original.
    """
    validator = jsonschema.Draft202012Validator(miagis_schema.metadata_schema)
    errors_generator = validator.iter_errors(metadata)
    
    for error in errors_generator:
        
        message = ""
        custom_message = ""
        
        if error.validator == "minProperties":
            custom_message = " cannot be empty."
        elif error.validator == "required":
            required_property = re.match(r"(\'.*\')", error.message).group(1)
            if len(error.relative_path) == 0:
                message += "The required property " + required_property + " is missing."
            else:
                message += "The entry " + "[%s]" % "][".join(repr(index) for index in error.relative_path) + " is missing the required property " + required_property + "."
        elif error.validator == "dependencies":
            message += "The entry " + "[%s]" % "][".join(repr(index) for index in error.relative_path) + " is missing a dependent property.\n"
            message += error.message
        elif error.validator == "dependentRequired":
            message += "The entry " + "[%s]" % "][".join(repr(index) for index in error.relative_path) + " is missing a dependent property.\n"
            message += error.message
        elif error.validator == "minLength":
            custom_message = " cannot be an empty string."
        elif error.validator == "maxLength":
            custom_message = " is too long."
        elif error.validator == "minItems":
            custom_message = " cannot be empty."
        elif error.validator == "type":
            if type(error.validator_value) == list:
                custom_message = " is not any of the allowed types: ["
                for allowed_type in error.validator_value:
                    custom_message += "\'" + allowed_type + "\', "
                custom_message = custom_message[:-2]
                custom_message += "]."
            else:
                custom_message = " is not of type \"" + error.validator_value + "\"."
        elif error.validator == "enum":
            custom_message = " is not one of [" + "%s" % ", ".join(repr(index) for index in error.validator_value) + "]."
        elif error.validator == "format":
            custom_message = " is not a valid " + error.validator_value + "."
        elif error.validator == "pattern":
            custom_message = " must be \"FAIR\", so it can only include the letters F, A, I, and R in that order, case-insensitive."
        elif error.validator == "minimum":
            custom_message = " must be greater than or equal to " + str(error.validator_value) + "."
        elif error.validator == "maximum":
            custom_message = " must be less than or equal to " + str(error.validator_value) + "."
        else:
            print(error.message)
        
        
        if custom_message:
            message = message + "The value for " + "[%s]" % "][".join(repr(index) for index in error.relative_path) + custom_message
        print(message)
        
        
    
    ## Check that all products are in resources.
    if "products" in metadata and "resources" in metadata:
        for product in metadata["products"]:
            if product not in metadata["resources"]:
                print("The product, " + product + ", is not in \"resources\".")
    
    ## Check that all sources exist in resources, schema are valid json schema, and field names match ids.
    if "resources" in metadata:
        for resource_name, resource_fields in metadata["resources"].items():
            
            if "sources" in resource_fields:
                for source in resource_fields["sources"]:
                    if not source in metadata["resources"]:
                        print("The source, " + source + ", for resource, " + resource_name + ", does not exist in resources.")
                
            if "schema" in resource_fields and type(resource_fields["schema"]) == dict:
                schema_validator = jsonschema.validators.validator_for(resource_fields["schema"])
                try:
                    schema_validator.check_schema(resource_fields["schema"])
                except jsonschema.SchemaError:
                    print("The \"schema\" field for resource, " + resource_name + ", is not a valid JSON Schema.")
                    
            if "fields" in resource_fields:
                for field_name, field_values in resource_fields["fields"].items():
                    if "name" in field_values and field_name != field_values["name"]:
                        print("The \"name\" property for field, " + field_name + ", for resource, " + resource_name + ", does not match its key value.")
            
            


# -*- coding: utf-8 -*-
"""
This module contains the code to build the metadata file.
"""

import pathlib
import datetime
import os
import copy
import typing as t

import fuzzywuzzy.fuzz
import pandas
import jsonschema

from . import miagis_schema
from . import user_input_checking


def build(resource_properties_path: str, exact_matching: bool =False, remove_optional_fields: bool =True, 
          add_resources: bool =True, overwrite_format: bool =False, overwrite_fairness: bool =False,
          base_metadata: dict ={}, entry_version: int =1, entry_id: str ="", base_description: str ="", products: list =[], schema_list: list = []) -> dict:
    """Build a metadata file from the input settings in the current directory.
    
    Loop over files in the folders of the current directory, ignoring files in 
    the top level of the current directory, and fill in what information can be 
    about the file in the metadata. If resource_properties are provided then resources 
    are matched to those in resource_properties and more information is able to be 
    automatically put into the metadata. 
    
    Args:
        resource_properties_path: filepath to a file containing file properties.
        exact_matching: if True file names are matched exactly. if False names are modified and matched fuzzy.
        remove_optional_fields: if True optional metadata fields that are empty or null are removed.
        add_resources: if True add resources that aren't in the metadata after looping over the files to the metadata.
        overwrite_format: if True overwrite the default format determined from the file extension for files with what is in resource_properties.
        overwrite_fairness: if True overwrite the default value of "FAIR" for files with what is in resource_properties.
        base_metadata: update the metadata dict with this input dict.
        entry_version: version number of the metadata.
        entry_id: unique id for the metadata.
        base_description: description of the metadata.
        products: a list of resource ids.
        schema_list: a list of dicitonaries with information about how to detect if a JSON file is using that schema, and how to decode that schema to fill in the metadata.
        
    Returns:
        The metadata JSON as a python dict.
    """
    directory = pathlib.Path.cwd()
    
    ## Add gejson and acrgis json to schema's to look for.
    arcgis_type_map = {"esriFieldTypeOID":"int", "esriFieldTypeString":"str", "esriFieldTypeInteger":"int", 
                       "esriFieldTypeSmallInteger":"int", "esriFieldTypeSingle":"float", "esriFieldTypeDouble":"float",
                       "esriFieldTypeSmallInteger":"int", "esriFieldTypeDate":"int", "typeIdField":"string"}
    
    schema_list.append({"name":"In-built ESRI", "style":"mapping", "schema":miagis_schema.arcgis_schema, 
                    "field_path":'["layers"][0]["layerDefinition"]["fields"]', 
                    "name_key":"name", "type_key":"type", "type_map":arcgis_type_map})
    schema_list.append({"name":"In-built GEOJSON Single Feature", "style":"testing", "schema":miagis_schema.geojson_feature_schema, 
                        "features_path":"", "properties_key":"properties", "schema_URL":"https://datatracker.ietf.org/doc/html/rfc7946"})
    schema_list.append({"name":"In-built GEOJSON Collection", "style":"testing", "schema":miagis_schema.geojson_collection_schema, 
                        "features_path":'["features"]', "properties_key":"properties", "schema_URL":"https://datatracker.ietf.org/doc/html/rfc7946"})
    
    ## Initially build the metadata, and then update it with base metadata, and then overwrite values with input arguments if not null.
    metadata = {
      "format_version" : "DRAFT_MIAGIS_VERSION_0.1", 
      "entry_version" : entry_version, 
      "entry_id" : entry_id,
      "date" : str(datetime.datetime.now().date()),
      "description" : base_description,
      "products" : products,
      "resources" : {}
    }
    
    metadata.update(base_metadata)
    if base_description:
        metadata["description"] = base_description
    if entry_version != 1:
        metadata["entry_version"] = entry_version
    if entry_id:
        metadata["entry_id"] = entry_id
    if products:
        metadata["products"] = products
    
    
    ## If --resource_properties was given read in the data.
    resource_properties, resource_original_name_map = user_input_checking.read_in_resource_properties(resource_properties_path, exact_matching)
    resource_properties_keys = list(resource_properties.keys())
    resource_matches = {}
    
    for root, directories, files in os.walk(directory):
        
        ## Only look in subfolders not the cwd.
        if pathlib.Path(root) == directory:
            continue
        
        relative_path = pathlib.Path(root).relative_to(directory).as_posix()
        folder_name = pathlib.Path(root).name
        
        for filename in files:
            extension = pathlib.Path(filename).suffix[1:].lower()
            filename_minus_extension = pathlib.Path(filename).stem
            
            relative_location = pathlib.Path(relative_path, filename).as_posix()
            
            if folder_name == "layer_data":
                file_type = "layer"
            elif folder_name == "map_data":
                file_type = "map"
            else:
                file_type = ""
                
            if extension == "py":
                extension = "python"
                file_type = "program"
                
            
            metadata["resources"][relative_location] = {"location":relative_location,
                                                    "type":file_type,
                                                    "description":"",
                                                    "fairness":"FAIR",
                                                    "format":extension,}
            
            ## Try to find properties for the file.
            current_resource_properties, alternate_locations, matched_name = find_resource_properties(resource_properties, resource_properties_keys, exact_matching, 
                                                                                                      filename_minus_extension, relative_location)
            
            ## Keep track of what files get matched so the alternate locations can all be updated at the end.
            if matched_name:
                if matched_name in resource_matches:
                    resource_matches[matched_name].append(relative_location)
                else:
                    resource_matches[matched_name] = [relative_location]
                
                metadata["resources"][relative_location].update(current_resource_properties)
                metadata["resources"][relative_location]["location"] = relative_location
                metadata["resources"][relative_location]["alternate_locations"] = alternate_locations
                
                if not overwrite_format:
                    metadata["resources"][relative_location]["format"] = extension
                if not overwrite_fairness:
                    metadata["resources"][relative_location]["fairness"] = "FAIR"
            
                        
            ## Determine what kind of file it is and attempt to fill in the types of the fields.
            if extension == "json" or extension == "geojson" or extension == "csv" or extension == "xlsx":
                
                path_to_read_file = pathlib.Path(root, filename)
                
                ## Fill in the fields for tabular file types.
                if extension == "csv" or extension == "xlsx":                        
                    metadata["resources"][relative_location]["fields"] = determine_table_fields(extension, path_to_read_file)
                
                else:
                    input_json = user_input_checking.load_json(path_to_read_file)
                    
                    json_fields, schema = determine_json_fields(schema_list, input_json, relative_location)
                    
                    if schema:
                        metadata["resources"][relative_location]["schema"] = schema
                        
                    metadata["resources"][relative_location]["fields"] = json_fields
            
            else:
                continue
            
    ## Add all resources to the metadata.
    if add_resources:
        for resource_name in resource_properties:
            if not resource_name in metadata["resources"] and \
                "location" in resource_properties[resource_name] and resource_properties[resource_name]["location"]:
                metadata["resources"][resource_original_name_map[resource_name]] = resource_properties[resource_name]
        
    ## Go through all matches and update alternate_locations so they all reference each other.
    for resource_name in resource_matches:
        total_list = resource_matches[resource_name]
        if resource_original_name_map[resource_name] in metadata["resources"]:
            total_list.append(resource_original_name_map[resource_name])
        
        all_locations = set()
        for name in total_list:
            all_locations.add(metadata["resources"][name]["location"])
            if "alternate_locations" in metadata["resources"][name]:
                for location in metadata["resources"][name]["alternate_locations"]:
                    all_locations.add(location)
        all_locations = list(all_locations)
        ## The sort is so testing is easier.
        all_locations.sort()
            
        for name in total_list:
            metadata["resources"][name]["alternate_locations"] = [location for location in all_locations if location != metadata["resources"][name]["location"]]
                    
                    
    ## Remove empty optional fields if the option was used.
    required_fields = miagis_schema.metadata_schema["properties"]["resources"]["additionalProperties"]["then"]["required"]
    if remove_optional_fields:
        for resource in metadata["resources"]:
            fields_to_delete = []
            for field, field_value in metadata["resources"][resource].items():
                if not field in required_fields and not field_value:
                    fields_to_delete.append(field)
            for field in fields_to_delete:
                del metadata["resources"][resource][field]
            
        
    return metadata
    
        



def find_resource_properties(resource_properties: dict, resource_properties_keys: list, exact_matching: bool, 
                         filename_minus_extension: str, relative_location: str) -> t.Tuple[dict, list, str]:
    """Match filename to an entry in resource_properties and pull out properties.
    
    The point of this function is to find the match in resource_properties and format 
    all of the information so it can be easily added to metadata without more logic 
    after returning from this function.
    
    Args:
        resource_properties: dictionary where the keys are filenames and values are properties of the file to go in the metadata.
        resource_properties_keys: list of the resource_properties_keys, it is given as an input so the list isn't created every time the function runs.
        exact_matching: if True filename is matched as is, if False the filename is stripped, lowered, and spaces replaced with underscores before matching
        filename_mins_extension: the filename to match with the extension removed.
        relative_location: location of the file relative to the current directory, input here so it can be added to alternate_locations.
        
    Returns:
        The properties of the matched filename from resource_properties.
        If alternate_locations is in resource_properties return it, else return an empty list, but always add the relative_location to the list.
        The resource_name in resource_properties that was matched to filename_minus_extension.
    """
    
    current_resource_properties = {}
    matched_name = ""
    if not exact_matching:
        fuzzy_match_filename = filename_minus_extension.strip()
        fuzzy_match_filename = fuzzy_match_filename.lower()
        fuzzy_match_filename = fuzzy_match_filename.replace(" ", "_")
        
        fuzzy_matches = [layer_name for layer_name in resource_properties_keys if fuzzywuzzy.fuzz.ratio(layer_name, fuzzy_match_filename) >= 90]
        if len(fuzzy_matches) == 1:
            current_resource_properties = resource_properties[fuzzy_matches[0]]
            matched_name = fuzzy_matches[0]
        elif len(fuzzy_matches) > 1:
            if fuzzy_match_filename in resource_properties:
                current_resource_properties = resource_properties[fuzzy_match_filename]
                matched_name = fuzzy_match_filename
                
    else:
        if filename_minus_extension in resource_properties:
            current_resource_properties = resource_properties[filename_minus_extension]
            matched_name = filename_minus_extension
           
    if "alternate_locations" in current_resource_properties:
        alternate_locations = copy.copy(current_resource_properties["alternate_locations"])
    else:
        alternate_locations = []
    if not relative_location in alternate_locations:
        alternate_locations.append(relative_location)
    if "location" in current_resource_properties and current_resource_properties["location"] and \
        not current_resource_properties["location"] in alternate_locations:
        alternate_locations.append(current_resource_properties["location"])
    
    ## Sort so testing is easier.
    alternate_locations.sort()
        
    return current_resource_properties, alternate_locations, matched_name



def determine_table_fields(extension: str, path_to_read_file: pathlib.Path) -> dict:
    """Read in the tabular file and determine the names and types of the columns.
    
    Args:
        extension: the extension of the file without the period, used to read in the file correctly.
        path_to_read_file: filepath to read in the file from.
        
    Returns:
        :Keys are the field name or column number and values are the name, type, and column number for each field. Ex: field_name : {"name":field_name, "type":field_type, "identifier":column_number+1, "identifier%type":"column"}.
    """
    
    if extension == "csv":
        df = pandas.read_csv(path_to_read_file, encoding_errors="ignore")
    else:
        df = pandas.read_excel(path_to_read_file)
        
    ## try to tell if the first row is a header or not.
    ## If there are any headers that are numbers assume there aren't headers.
    has_headers = True
    for header in df.columns:
        try:
            float(header)
            has_headers = False
        except ValueError:
            pass
    
    
    fields_dict = df.dtypes.to_dict()
    
    for field, dtype in fields_dict.items():
        dtype_as_string = str(dtype)
        if "int" in dtype_as_string:
            new_dtype = "int"
        elif "float" in dtype_as_string:
            new_dtype = "float"
        else:
            new_dtype = "str"
            
        fields_dict[field] = new_dtype
    
    fields = {}
    if has_headers:
        for field_name, field_type in fields_dict.items():
            if not df.loc[:, field_name].isna().all():
                fields[field_name] = {"name":field_name, "type":field_type, "identifier":df.columns.get_loc(field_name)+1, "identifier%type":"column"}
    else:
        for field_name, field_type in fields_dict.items():
            if not df.loc[:, field_name].isna().all():
                column_number = df.columns.get_loc(field_name)
                fields[column_number+1] = {"name":column_number+1, "type":field_type, "identifier":column_number+1, "identifier%type":"column"}
        
    return fields



def determine_json_fields(schema_list: list, input_json: dict, file_path: str) -> t.Tuple[dict,t.Union[dict,str]]:
    """Determine the types of the feature properties in the JSON file.
    
    Based on the ESRIJSON and GEOJSON formats there are 2 types of schema styles, 
    mapping and testing. The mapping style has types already identified with the 
    properties in the JSON, but the types are not the same as those defined in the 
    MIAGIS schema, so a mapping between the types is necessary. This style's dict 
    entry looks like:
        
        {"style":"mapping", "schema":valid_jsonschema, "field_path":path_to_fields, 
         "name_key":key_to_name, "type_key":key_to_type, "type_map":type_mapping}
    
    Actual ESRIJSON example:
        {"style":"mapping", "schema":miagis_schema.arcgis_schema, 
         "field_path":'["layers"][0]["layerDefinition"]["fields"]', 
         "name_key":"name", "type_key":"type", "type_map":arcgis_type_map}
    
    The testing style doesn't have types listed directly in the JSON so each field's 
    type must be tested. This style's dict entry looks like:
        
        {"style":"testing", "schema":valid_jsonschema, 
         "features_path":path_to_features, "properties_key":key_to_properties}
        
    Actual GEOJSON example:
        {"style":"testing", "schema":miagis_schema.geojson_feature_schema, 
         "features_path":"", "properties_key":"properties", "schema_URL":"https://datatracker.ietf.org/doc/html/rfc7946"}
        
    Note that the features_path is assumed to lead to a dict or list, and will 
    print a warning if does is not.
        
    Args:
        schema_list: list of dictionaries where each dictionary describes a schema to look for and how to determine field types.
        input_json: the JSON to determine field types for.
        file_path: path to the JSON file, used for printing error messages.
        
    Returns:
        :A dictionary of the fields and their types. Ex field_name : {"name":field_name, "type":field_type}.
        If "schema_URL" is in the schema dictionary then this is returned, otherwise return the jsonschema.
    """
    
    schema_properties = {}
    for format_properties in schema_list:
        
        try:
            jsonschema.validate(input_json, format_properties["schema"])
            schema_properties = format_properties
            break
        except jsonschema.ValidationError:
            continue
        
    if not schema_properties:
        return {}, ""
        
        
    if "schema_URL" in schema_properties:
        schema = schema_properties["schema_URL"]
    else:
        schema = schema_properties["schema"]
        
    if schema_properties["style"] == "mapping":
        
        field_path = schema_properties["field_path"]
        try:
            fields = eval("input_json" + field_path)
        except Exception:
            print("Warning: The \"field_path\", " + field_path + ", for the \"" + schema_properties["name"] + 
                  "\" json format schema does not work for file " + file_path + 
                  ". It will have an empty \"fields\" in the output.")
            return {}, schema
        
        type_key = schema_properties["type_key"]
        name_key = schema_properties["name_key"]
        type_map = schema_properties["type_map"]
            
        json_fields = {}
        type_already_printed = False
        name_already_printed = False
        for field in fields:
            if not type_key in field:
                if not type_already_printed:
                    print("Warning: The \"type_key\", " + type_key + ", for the \"" + schema_properties["name"] + 
                          "\" json format schema is not in all of the fields for file " + 
                          file_path + ". Some fields may be missing in the output.")
                    type_already_printed = True
                continue
            
            if not name_key in field:
                if not name_already_printed:
                    print("Warning: The \"name_key\", " + name_key + ", for the \"" + schema_properties["name"] + 
                          "\" json format schema is not in all of the fields for file " + 
                          file_path + ". Some fields may be missing in the output.")
                    name_already_printed = True
                continue
                    
            if field[type_key] in type_map:
                json_fields[field[name_key]] = {"name":field[name_key], "type":type_map[field[type_key]]}
            else:
                json_fields[field[name_key]] = {"name":field[name_key], "type":"UNKNOWN"}
                
        return json_fields, schema
        
    
    elif schema_properties["style"] == "testing":
        
        features_path = schema_properties["features_path"]
        try:
            features = eval("input_json" + features_path)
        except Exception:
            print("Warning: The \"features_path\", " + features_path + " for the \"" + schema_properties["name"] + 
                  "\" json format schema does not work for file " + file_path + 
                  ". It will have empty \"fields\" in the output.")
            return {}, schema
        
        if type(features) != list and type(features) != dict:
            print("Warning: The \"features_path\", " + features_path + " for the \"" + schema_properties["name"] + 
                  "\" json format schema does not lead to the appropriate type (list or dict) for file " + file_path 
                  + ". It will have empty \"fields\" in the output.")
            return {}, schema
        
        ## In the case where there is only a single feature we make the dictionary become a list with that dictionary inside it.
        if type(features) == dict:
            features = [features]
        
        properties_key = schema_properties["properties_key"]
        
        ## This loop assumes that each feature has the same properties and tries to determine the type of each property.
        ## It does not enforce features to have all the same properties. It simply accumulates all features it sees.
        ## Also assumes that any property will only have one type, so the first type seen is kept and not checked or overwritten except for None.
        json_fields = {}
        properties_already_printed = False
        for feature in features:
                        
            if not properties_key in feature:
                if not properties_already_printed:
                    print("Warning: The \"properties_key\", " + properties_key + " for the \"" + schema_properties["name"] + 
                          "\" json format schema is not in all of the features for file " + 
                          file_path + ". Some fields may be missing in the output.")
                    properties_already_printed = True
                continue
            
            if feature[properties_key]:
                for key, value in feature[properties_key].items():
                    value_type = type(value)
                    if value_type == int:
                        str_type = "int"
                    elif value_type == float:
                        str_type = "float"
                    elif value_type == str:
                        str_type = "str"
                    elif value is None:
                        str_type = "None"
                    else:
                        str_type = str(value_type)
                        
                    if key in json_fields and json_fields[key]["type"] != "None":
                        continue
                    else:
                        json_fields[key] = {"name":key, "type":str_type}
    
        ## Remove any fields without a type.
        fields_to_delete = []
        for field_name, field_properties in json_fields.items():
            if field_properties["type"] == "None":
                fields_to_delete.append(field_name)
                
        for field_name in fields_to_delete:
            del json_fields[field_name]
            
        return json_fields, schema
        
    else:
        print("Warning: Unknown \"style\", " + schema_properties["style"] + " for the \"" + schema_properties["name"] + 
              "\" json format. The file at " + file_path 
              + " will have empty \"fields\" in the output.")
        return {}, schema
            
            
        





# -*- coding: utf-8 -*-
"""
This module contains the code to print and save map structure.
"""

import pathlib

def print_map_layers(metadata: dict, save_path: str =""):
    """Find the maps and layers in metadata and pretty print them.
    
    Args:
        metadata: A valid metadata to find maps in.
        save_path: path to save the output to a file, doesn't save if it is an empty string.
    """
    
    maps_dict = {}
    for resource_name, resource_properties in metadata["resources"].items():
        if resource_properties["type"] == "map":
            layers = []
            if "sources" in resource_properties:
                for source in resource_properties["sources"]:
                    if metadata["resources"][source]["type"] == "layer":
                        layers.append(source)
            maps_dict[resource_name] = layers
    
    print_string = "Maps:\n"
    for map_name, layers in maps_dict.items():
        print_string += "\t" + map_name + "\n"
        print_string += "\t\tLayers:\n"
        for layer in layers:
            print_string += "\t\t\t" + layer + "\n"
        print_string += "\n"
            
    print(print_string)
    
    if save_path:
        extension = pathlib.Path(save_path).suffix[1:].lower()
        if extension != "txt":
            save_path += ".txt"
        with open(save_path, 'wb') as outFile:
            outFile.write(print_string.encode("utf-8"))


# -*- coding: utf-8 -*-
"""
This module contains the functions used to read in and validate various inputs.
"""

import sys
import pathlib
import json
import re
import typing as t

import jsonschema
import pandas



def load_json(filepath: str) -> dict:
    """Adds error checking around loading a json file.
    
    Args:
        filepath: filepath to the json file
        
    Returns:
        json read from file in a dictionary
        
    Raises:
        Exception: If file opening has a problem will raise an exception.
    """
    if pathlib.Path(filepath).exists():
        try:
            with open(filepath, "r") as f:
                internal_data = json.loads(f.read())
        except Exception as e:
            raise e

        return internal_data
    else:
        print("No such file: " + filepath)
        sys.exit()



def read_in_resource_properties(resource_properties_path: str, exact_matching: bool) -> t.Tuple[dict,dict]:
    """Read in resource_properties and put it in expected dict form.
    
    resource_properties can be csv, xlsx, or JSON, so if it is one of the tabular 
    forms some of the fields have to be read in special.
    
    Args:
        resource_properties_path: filepath to the resource_properties file.
        exact_matching: if True resource names will not be modified. 
                               if False resource names are stripped, lowered, and spaces replaced with underscores.
    
    Returns:
        The final dictionary of resource_properties, and a mapping of the original resource name and new name.
    """
    
    if resource_properties_path == None:
        return {}, {}
    
    if not pathlib.Path(resource_properties_path).exists():
        print("No such file: " + resource_properties_path)
        sys.exit()
    
    extension = pathlib.Path(resource_properties_path).suffix[1:].lower()
    if extension == "csv":
        resource_properties_df = pandas.read_csv(resource_properties_path, dtype=str)
    elif extension == "xlsx":
        resource_properties_df = pandas.read_excel(resource_properties_path, dtype=str)
    elif extension == "json":
        resource_properties = load_json(resource_properties_path)
        if not exact_matching:
            resource_original_name_map = {key.strip().lower().replace(" ", "_"):key for key in resource_properties}
            resource_properties = {key.strip().lower().replace(" ", "_"):value for key, value in resource_properties.items()}
        else:
            resource_original_name_map = {key:key for key in resource_properties}
        return resource_properties, resource_original_name_map
    else:
        print("Error: Unknown file type for --resource_properties.")
        sys.exit()
    
    if not "resource_name" in resource_properties_df.columns:
        print("Error: The file input for --resource_properties does not have a resource_name column.")
        sys.exit()
    
    resource_properties_df = resource_properties_df.fillna("")
        
    if "alternate_locations" in resource_properties_df:
        resource_properties_df.loc[:, "alternate_locations"] = resource_properties_df.loc[:, "alternate_locations"].str.strip()
    
    resource_properties_df = resource_properties_df.drop_duplicates()
    resource_properties_df = resource_properties_df.set_index("resource_name", drop=True)
    
    resource_properties = resource_properties_df.to_dict(orient="index")
    
    if not exact_matching:
        resource_original_name_map = {key.strip().lower().replace(" ", "_"):key for key in resource_properties}
        resource_properties = {key.strip().lower().replace(" ", "_"):value for key, value in resource_properties.items()}
    else:
        resource_original_name_map = {key:key for key in resource_properties}
    
    for resource_name, properties in resource_properties.items():
        if "alternate_locations" in resource_properties_df.columns:
            resource_properties[resource_name]["alternate_locations"] = [location.strip() for location in properties["alternate_locations"].split(",") if location.strip()]
            
        if "sources" in resource_properties_df.columns:
            resource_properties[resource_name]["sources"] = [source.strip() for source in properties["sources"].split(",") if source.strip()]
            
        if "creator" in resource_properties_df.columns and "creator_type" in resource_properties_df.columns:
            creators = [creator.strip() for creator in properties["creator"].split(",")]
            creator_types = [creator_type.strip() for creator_type in properties["creator_type"].split(",") if creator_type]
            if len(creators) == len(creator_types):
                resource_properties[resource_name]["creator"] = [{"name":creators[i], "type":creator_types[i]} for i in range(len(creators))]
            else:
                print("Warning: Not every creator in \"creator\" has a \"creator_type\" for " + resource_name + " in the --resource_properties file.")
            
            del[resource_properties[resource_name]["creator_type"]]
    
    
    return resource_properties, resource_original_name_map




def validate_arbitrary_schema(dict_to_validate: dict, schema: dict):
    """Validate any arbitrary JSON Schema.
    
    Wraps around jsonschema.validate() to give more human readable errors 
    for most validation errors.
    
    Args:
        dict_to_validate: instance to validate.
        schema: JSON Schema to validate the instance with.
        
    Raises:
        jsonschema.ValidationError: any validation errors that aren't handled reraise the original.
    """
        
    try:
        jsonschema.validate(dict_to_validate, schema)
    except jsonschema.ValidationError as e:
        
        message = "ValidationError: An error was found in the " + schema["title"] + ".\n"
        custom_message = ""
        
        if e.validator == "minProperties":
            message += "The entry " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + " cannot be empty."
        elif e.validator == "required":
            required_property = re.match(r"(\'.*\')", e.message).group(1)
            if len(e.relative_path) == 0:
                message += "The required property " + required_property + " is missing."
            else:
                message += "The entry " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + " is missing the required property " + required_property + "."
        elif e.validator == "minLength":
            custom_message = " cannot be an empty string."
        elif e.validator == "maxLength":
            custom_message = " is too long."
        elif e.validator == "minItems":
            custom_message = " cannot be empty."
        elif e.validator == "type":
            if type(e.validator_value) == list:
                custom_message = " is not any of the allowed types: ["
                for allowed_type in e.validator_value:
                    custom_message += "\'" + allowed_type + "\', "
                custom_message = custom_message[:-2]
                custom_message += "]."
            else:
                custom_message = " is not of type \"" + e.validator_value + "\"."
        elif e.validator == "enum":
            custom_message = " is not one of [" + "%s" % ", ".join(repr(index) for index in e.validator_value) + "]."
        elif e.validator == "format":
            custom_message = " is not a valid " + e.validator_value + "."
        elif e.validator == "minimum":
            custom_message = " must be greater than or equal to " + str(e.validator_value) + "."
        elif e.validator == "maximum":
            custom_message = " must be less than or equal to " + str(e.validator_value) + "."
        else:
            raise e
        
        
        if custom_message:
            message = message + "The value for " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + custom_message
        print(message)
        sys.exit()


def additional_args_checks(args: dict):
    """Run some checks on args that jsonschema can't do.
    
    This assumes that args has been validated with a JSON schema and does some 
    further checking to make sure the values entered by the user make sense. 
    Prints a message and exits the program if problems are found.
    
    Args:
        args: the arguments entered into the program by the user.
    """
    file_path_properties = ["--resource_properties", "--base_metadata", "--json_schemas", "<metadata_json>"]
    for path in file_path_properties:
        if args[path] and not pathlib.Path(args[path]).exists():
            print("Error: The value entered for " + path + " is not a valid file path or does not exist.")
            sys.exit()
                
        
    try:
        entry_version = int(args["--entry_version"])
    except ValueError:
        print("Error: The value entered for --entry_version is not an integer.")
        sys.exit()
        
    if entry_version < 1:
        print("Error: The value entered for --entry_version is less than 1.")
        sys.exit()
        

def additional_json_schemas_checks(schema_list: list):
    """Check that all input schemas are valid JSON Schema.
    
    If any schema are not valid JSON Schema then print a message and exit.
    
    Args:
        schema_list: list of dictionaries of properties for JSON schemas.
    """
    
    ## Check that each schema is valid jsonschema.
    for i, format_properties in enumerate(schema_list):
        schema = format_properties["schema"]
        validator = jsonschema.validators.validator_for(schema)
        try:
            validator.check_schema(schema)
        except jsonschema.SchemaError:
            print("Error: The schema for index " + str(i) + " in the input JSON schema list is not valid JSON Schema.")
            sys.exit()
    
    
    



"""

This package has the following modules:
    
``build``
    This module contains functions for building the metadata file.

``miagis_schema``
    This module contains the schema for validating user input.
    
``validate``
    This module contains functions to validate the metadata file.
    
``user_input_checking``
    This module contains functions to validate user input.
    
``print_map_layers``
    This module contains the code to run the print_map_layers command.
    

"""
__version__ = "1.0.1"


# -*- coding: utf-8 -*-
"""
    GitHub: https://github.com/MoseleyBioinformaticsLab/miagis/tree/main
    Documentation: https://moseleybioinformaticslab.github.io/miagis/

    For the current directory go through all files and folders and build a GIS metadata file.
    The output is saved as GIS_METADATA.json in the current directory and will overwrite without warning.
    
    The output is simply an attempt to automate as much of the metadata build as possible, 
    and is not guarenteed to be correct or be a valid metadata file. It is expected that 
    after the initial build a user will double check and correct by hand, and then use 
    the validate command to validate their metadata before submission.
    
    The resource_properties option file is expected to have a header row on the first row for tabular 
    files with "file_name" being the only required row to use as a key. Other headers should 
    match field names to go into the metadata[files] properties. Extra property names beyond 
    those in the metadata specification are allowed and simply added as is to any items.
    
    Example:
        file_name	    alternate_locations	     sources	        source_types	    description	           geographical_area
        example_name	URL_to_file	             source1,source2	organization,URL	example_description	   Kentucky

    
    Any directory structure will work, but there is additional functionality if the directories 
    are named a certain way. If layer data is kept in "layer_data" then the program will add them 
    to metadata[products][layers]. The same for maps in "map_data". 
    
    The standard example run:
        miagis build --remove_optional_fields --resource_properties <filepath> --base_metadata <filepath>
    
 
 Usage:
    miagis build [options]
    miagis validate <metadata_json>
    miagis print_map_layers <metadata_json> [--save_path=<save_path>]

 Options:
    --help                              Show this help documentation.
    --resource_properties=<file_path>   Filepath to a csv, xlsx, or JSON file with file properties.
    --exact_name_match                  If used then file name matching will be done exactly instead of fuzzy.
    --add_resources                     If used then add resources from resource_properties directly to the metadata.
    --overwrite_format                  If used then overwrite the determined format for files with what is in resource_properties.
    --overwrite_fairness                If used then overwrite the determined fairness for files with what is in resource_properties.
    --remove_optional_fields            If used then delete optional metadata fields that are empty from files.
    --json_schemas=<file_path>          Filepath to a JSON file with schemas for different JSON formats.
    

 Base Metadata Options:
    --entry_version=<integer>           Set the entry_version field for the metadata. Should be an integer starting from 1. [default: 1]
    --entry_id=<id>                     Set the entry_id field for the metadata.
    --description=<description>         Set the description field for the metadata
    --base_metadata=<file_path>         Filepath to a JSON file with the base metadata fields to use.
"""


import warnings
import pathlib
import json

warnings.filterwarnings("ignore", module="fuzzywuzzy")

import docopt

from . import miagis_schema
from . import build
from . import validate
from . import user_input_checking
from . import print_map_layers



def main():
    args = docopt.docopt(__doc__)
    user_input_checking.validate_arbitrary_schema(args, miagis_schema.args_schema)
    user_input_checking.additional_args_checks(args)
        
    if args["build"]:
        
        if args["--json_schemas"]:
            schema_list = user_input_checking.load_json(args["--json_schemas"])
            user_input_checking.validate_arbitrary_schema(schema_list, miagis_schema.schema_list_schema)
            user_input_checking.additional_json_schemas_checks(schema_list)
        else:
            schema_list = []
        
        if args["--base_metadata"]:
            base_metadata = user_input_checking.load_json(args["--base_metadata"])
        else:
            base_metadata = {}
        
        
        metadata = build.build(args["--resource_properties"], args["--exact_name_match"], 
                    args["--remove_optional_fields"], args["--add_resources"],
                    args["--overwrite_format"], args["--overwrite_fairness"],
                    base_metadata, int(args["--entry_version"]), args["--entry_id"], 
                    args["--description"], [], schema_list)
        
        save_path = pathlib.Path(pathlib.Path.cwd(), "GIS_METADATA.json")
        with open(save_path, 'w') as outFile:
            print(json.dumps(metadata, indent=2, sort_keys=False), file=outFile)
    
    elif args["validate"]:
        validate.validate(user_input_checking.load_json(args["<metadata_json>"]))
    elif args["print_map_layers"]:
        print_map_layers.print_map_layers(user_input_checking.load_json(args["<metadata_json>"]), args["--save_path"])
    else:
        print("Unrecognized command")






if __name__ == "__main__":
    main()



